// sldemo.c

// Screensaver library demonstration program.

#include <windows.h>

#include "sldemo.h"
#include "slib.h"

#pragma resource "sldemo.res"

// configuration data -- type of shape to draw and maximum number of shapes that
// can be drawn before clearing the screen

static enum {shCircle, shRectangle} g_shape = shCircle;

static int g_iMaxDrawnShapes = 50;

// local function prototypes

void ReadConfig  (void);
void WriteConfig (void);

// =========================================================================
// BOOL WINAPI RegisterDialogClasses (HANDLE hinstance)
//
// Allow the programmer to register any child control windows.
//
// Arguments:
//
// hinstance - screensaver instance handle
//
// Return:
//
// TRUE if child control windows successfully created or no controls needed;
// otherwise FALSE
// =========================================================================

#pragma argsused
BOOL WINAPI RegisterDialogClasses (HANDLE hinstance)
{
   return TRUE; // no controls need to be registered.
}

// ===========================================================================
// BOOL WINAPI ScreenSaverConfigureDialog (HWND hwnd, int iMsg, WPARAM wparam,
//                                         LPARAM lparam)
//
// Screensaver-specific configuration procedure for processing messages.
//
// Arguments:
//
// hwnd   - handle of window associated with this screensaver procedure
//
// iMsg   - message identifier
//
// wparam - 32-bit word parameter with message data (if applicable)
//
// lparam - 32-bit long word parameter with message data (if applicable)
//
// Return:
//
// TRUE if message processed; otherwise FALSE if message not processed or
// Windows should not set the default keyboard focus (WM_INITDIALOG)
// ===========================================================================

#pragma argsused
BOOL WINAPI ScreenSaverConfigureDialog (HWND hwnd, UINT iMsg, WPARAM wparam,
                                        LPARAM lparam)
{
   char szBuffer [256];

   switch (iMsg)
   {
      case WM_INITDIALOG:
           // Read configuration data from Windows registry.

           ReadConfig ();

           // Update dialog box controls based on configuration data.

           SendDlgItemMessage (hwnd, (g_shape == shCircle) ? IDR_CIRCLES
                                                           : IDR_RECTANGLES,
                               BM_SETCHECK, 1, 0);

           wsprintf (szBuffer, "%d", g_iMaxDrawnShapes);
           SetWindowText (GetDlgItem (hwnd, IDE_MDS), szBuffer);
           break;

      case WM_COMMAND:
           if (wparam == IDOK)
           {
               // Fetch configuration data from dialog box controls.

               if (!SendDlgItemMessage (hwnd, IDR_CIRCLES, BM_GETCHECK, 0, 0))
                   g_shape = shRectangle;
               else
                   g_shape = shCircle;

               if (GetWindowText (GetDlgItem (hwnd, IDE_MDS), szBuffer, 256))
                   g_iMaxDrawnShapes = atoi (szBuffer);

               // Write configuration data to Windows registry.

               WriteConfig ();
           }

           if (wparam == IDOK || wparam == IDCANCEL)
           {
               // Cancel dialog box.

               EndDialog (hwnd, 0);

               return TRUE;
           }
   }

   return FALSE;
}

// =====================================================================
// LRESULT WINAPI ScreenSaverProc (HWND hwnd, int iMsg, WPARAM wparam,
//                                 LPARAM lparam)
//
// Screensaver-specific procedure for processing messages.
//
// Arguments:
//
// hwnd   - handle of window associated with this screensaver procedure
//
// iMsg   - message identifier
//
// wparam - 32-bit word parameter with message data (if applicable)
//
// lparam - 32-bit long word parameter with message data (if applicable)
//
// Return:
//
// 0 if message processed, otherwise DefScreenSaverProc() result
// =====================================================================

LRESULT WINAPI ScreenSaverProc (HWND hwnd, UINT iMsg, WPARAM wparam,
                                LPARAM lparam)
{
   HDC hdc;
   RECT rect;
   int iWidth, iX, iY;
   HBRUSH hbrush, hbrushOld;

   static int iNumDrawnShapes;

   static UINT uTimerID;

   switch (iMsg)
   {
      case WM_CREATE:
           // Create an animation timer that results in a WM_TIMER message
           // approximately every 50 milliseconds.

           uTimerID = SetTimer (hwnd, 1, 50, NULL);

           // Read configuration data from Windows registry.

           ReadConfig ();

           return 0;

      case WM_DESTROY:
           // Destroy the timer.

           KillTimer (hwnd, uTimerID);

           // Inform Windows that screensaver is terminating.

           PostQuitMessage (0);

           return 0;

      case WM_TIMER:
           // Grab a device context associated with this window.

           hdc = GetDC (hwnd);

           // Obtain window coordinates.

           GetClientRect (hwnd, &rect);

           // Create a solid brush with a random color.

           hbrush = CreateSolidBrush (RGB (rand () % 256, rand () % 256,
                                      rand () % 256));

           // Select solid brush into device context so that filled shape
           // appears in brush color.

           hbrushOld = SelectObject (hdc, hbrush);

           // Compute upper-left corner coordinates.

           iX = rand () % rect.right;
           iY = rand () % rect.bottom;

           // Compute width of shape -- 10% of the lesser of the window's width
           // and height.

           iWidth = rect.right / 10;
           if (rect.bottom / 10 < iWidth)
               iWidth = rect.bottom / 10;

           // Draw either a filled ellipse or a filled rectangle.

           if (g_shape == shCircle)
               Ellipse (hdc, iX, iY, iX+iWidth, iY+iWidth);
           else
               Rectangle (hdc, iX, iY, iX+iWidth, iY+iWidth);

           // Increment drawn shape counter. If counter reaches the limit,
           // erase the screen and reset counter to 0.

           iNumDrawnShapes++;
           if (iNumDrawnShapes == g_iMaxDrawnShapes)
           {
               InvalidateRect (hwnd, NULL, TRUE);
               iNumDrawnShapes = 0;
           }

           // Select original brush into the device context and delete created
           // solid brush.

           SelectObject (hdc, hbrushOld);
           DeleteObject (hbrush);

           // Release the device context.

           ReleaseDC (hwnd, hdc);

           return 0;
   }

   return DefScreenSaverProc (hwnd, iMsg, wparam, lparam);
}

// ====================================================
// void ReadConfig (void)
//
// Read configuation information from Windows registry.
//
// Arguments:
//
// none
//
// Return:
//
// none
// ====================================================

static void ReadConfig (void)
{
   LONG res;
   HKEY skey;
   DWORD val, valSize, valType;

   // Open the specified key.

   res = RegOpenKeyEx (HKEY_CURRENT_USER, "Software\\Jeff Friesen\\SLDEMO", 0,
                       KEY_ALL_ACCESS, &skey);

   if (res == ERROR_SUCCESS)
   {
       // Fetch shape value from the Shape field of the open key.

       valSize = sizeof(val);
       res = RegQueryValueEx (skey, "Shape", 0, &valType, (LPBYTE)&val,
                              &valSize);

       if (res==ERROR_SUCCESS) g_shape = val;

       // Fetch maximum number of drawn shapes value from the MDS field of the
       // open key.

       valSize = sizeof(val);
       res = RegQueryValueEx (skey, "MDS", 0, &valType, (LPBYTE)&val,
                              &valSize);

       if (res==ERROR_SUCCESS) g_iMaxDrawnShapes = val;

       // Close the open key.

       RegCloseKey (skey);
   }
}

// ===================================================
// void WriteConfig (void)
//
// Write configuation information to Windows registry.
//
// Arguments:
//
// none
//
// Return:
//
// none
// ===================================================

static void WriteConfig (void)
{
   LONG res;
   HKEY skey;
   DWORD disp, val;

   // Create the specified key if it does not exist or open the key if it does
   // exist.

   res = RegCreateKeyEx (HKEY_CURRENT_USER, "Software\\Jeff Friesen\\SLDEMO", 0,
                         NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
                         &skey, &disp);

   if (res == ERROR_SUCCESS)
   {
       // Store shape value in the Shape field of the open key.

       val = g_shape;
       RegSetValueEx (skey, "Shape", 0, REG_DWORD, (CONST BYTE *) &val,
                      sizeof(val));

       // Store maximum number of drawn shapes value in the MDS field of the
       // open key.

       val = g_iMaxDrawnShapes;
       RegSetValueEx (skey, "MDS", 0, REG_DWORD, (CONST BYTE *) &val,
                      sizeof(val));

       // Close the open key.

       RegCloseKey (skey);
   }
}
